"""
contains class Callback
"""


class Callback(dict):
    """Callback passes parameters from earlier API calls back to the server"""

    def __init__(self, handle, callback):

        # derive from dict so instances get auto serialized to JSON
        # when passed as a parameter to an API call
        super().__init__(handle=handle, callback=callback)



"""exception handling routines"""

import logging as _logging


def handle_exception(exception):
    """logs an exceptions and returns an error message and error code"""

    if isinstance(exception, ApiError):
        _logging.exception(exception.message)
        return exception.message, exception.statuscode

    _logging.exception(repr(exception))
    return None, 500


class ApiError(Exception):

    """base class of exceptions thrown by the fdrtd API"""

    def __init__(self, statuscode, message):
        super().__init__()

        self.statuscode = statuscode
        self.message = message

    def __str__(self):
        return self.message


class InternalServerError(ApiError):
    """exception child class: internal server error"""

    def __init__(self, message):
        super().__init__(500, f'internal server error: {message}')


class NotAvailable(ApiError):
    """exception child class: not implemented / not available"""

    def __init__(self, missing):
        super().__init__(501, f'not implemented / not available: {missing}')


class MissingParameter(ApiError):
    """exception child class: missing parameter"""

    def __init__(self, missing):
        super().__init__(400, f'missing parameter: {missing}')


class InvalidParameter(ApiError):
    """exception child class: invalid parameter"""

    def __init__(self, parameter, invalid):
        super().__init__(400, f'invalid parameter: {parameter} = {invalid}')


class InvalidIdentifier(ApiError):
    """exception child class: invalid identifier"""

    def __init__(self, identifier, invalid):
        super().__init__(404, f'invalid identifier: {identifier} = {invalid}')


class MicroserviceNotFound(ApiError):
    """this exception is thrown when a microservice does not exist on the bus"""

    def __init__(self, missing):
        super().__init__(404, f'microservice not available: {missing}')


class FunctionNotFound(ApiError):
    """this exception is thrown when a member function of a microservice
       does not exist"""

    def __init__(self, missing):
        super().__init__(404, f'function not available: {missing}')


class FunctionNotPublic(ApiError):
    """this exception is thrown when a member function of a microservice
       exists, but is not public"""

    def __init__(self, missing):
        super().__init__(403, f'function not public: {missing}')



#!/usr/bin/env python3

"""
the main module
"""

import sys
import connexion
import waitress

from flask import current_app

from fdrtd.server.bus import Bus
from fdrtd.server.discovery import discover_microservices
from fdrtd.webserver.connexion.encoder import JSONEncoder


def main():
    """runs the webserver"""

    # enable the following line to log endpoint traffic
    # logging.basicConfig(level=logging.INFO)

    port = 8080
    for arg in sys.argv:
        if arg[:7] == "--port=":
            port = int(arg[7:])

    bus = Bus()
    microservices, classes = discover_microservices(bus)
    bus.set_microservices(microservices)
    bus.set_classes(classes)

    app = connexion.App(__name__, specification_dir='openapi/')
    app.app.json_encoder = JSONEncoder
    app.add_api('openapi.yaml',
                arguments={'title': 'fdrtd'},
                pythonic_params=True)

    with app.app.app_context():
        current_app.bus = bus

    @app.app.after_request
    def after_request(response):
        response.headers.add('Access-Control-Allow-Origin', '*')
        response.headers.add('Access-Control-Allow-Headers', 'Content-Type,Authorization')
        response.headers.add('Access-Control-Allow-Methods', 'GET,PUT,POST,PATCH,DELETE')
        return response

    waitress.serve(app, host="0.0.0.0", port=port)


if __name__ == '__main__':
    main()



"""
contains microservice KeyValueStorage
"""

import uuid as _uuid

from fdrtd.server.microservice import Microservice


class KeyValueStorage(Microservice):
    """stores and retrieves values by key"""

    def __init__(self, bus, endpoint):
        super().__init__(bus, endpoint)
        self.storages = {'default': {}}

    def create(self, value, storage='default'):
        """create a storage; store value, return key"""
        if storage not in self.storages:
            self.storages[storage] = {}
        uuid = str(_uuid.uuid4())
        callback = {'uuid': uuid, 'storage': storage}
        self.store(value, callback)
        return self.callback(callback)

    def store(self, value, callback):
        """store value, return key"""
        kvstorage = self.storages[callback['storage']]
        kvstorage[callback['uuid']] = value

    def retrieve(self, callback):
        """retrieve value from storage"""
        kvstorage = self.storages[callback['storage']]
        value = kvstorage[callback['uuid']]
        return value

    def exists(self, callback):
        """return true if key exists in storage"""
        kvstorage = self.storages[callback['storage']]
        return callback['uuid'] in kvstorage

    def delete(self, callback):
        """delete key from storage"""
        kvstorage = self.storages[callback['storage']]
        del kvstorage[callback['uuid']]



from fdrtd.builtins.sync.barrier import Barrier
from fdrtd.builtins.sync.broadcast import Broadcast


def get_microservices():
    return [
        {
            "identifiers": {
                "namespace": "fdrtd",
                "plugin": "Sync",
                "version": "0.3.0",
                "microservice": "Barrier"
            },
            "class": Barrier,
            "public": ["create", "delete"]
        },
        {
            "identifiers": {
                "namespace": "fdrtd",
                "plugin": "Sync",
                "version": "0.3.0",
                "microservice": "Broadcast"
            },
            "class": Broadcast,
            "public": ["create", "send", "receive", "delete"]
        }
    ]



from fdrtd.builtins.util.kvstorage import KeyValueStorage


def get_microservices():
    return [
        {
            "identifiers": {
                "namespace": "fdrtd",
                "plugin": "Util",
                "version": "0.3.0",
                "microservice": "KeyValueStorage"
            },
            "class": KeyValueStorage,
            "public": ["create", "store", "retrieve", "exists", "delete"]
        }
    ]



from fdrtd.server.bus import Bus
from fdrtd.server.discovery import discover_microservices

fdrtd_bus = Bus()
_microservices, _classes = discover_microservices(fdrtd_bus)
fdrtd_bus.set_microservices(_microservices)
fdrtd_bus.set_classes(_classes)




"""
contains class Microservice
"""

import fdrtd.server.exceptions
from fdrtd.server.callback import Callback


class Microservice:
    """base class for microservices"""

    def __init__(self, bus, handle):
        self.bus = bus
        self.handle = handle

    def callback(self, callback):
        """create a callback (used by child classes)"""
        return Callback(self.handle, callback)

    @staticmethod
    def make_public():
        """report public functions (will become obsolete!)
           note that child classes will overload this with make_public(self)"""
        return {}

    @staticmethod
    def safe_params(params, key):
        """helper routine, throws exception if key is not present in params"""
        if key not in params:
            raise fdrtd.server.exceptions.MissingParameter(key)
        return params[key]

    @staticmethod
    def safe_get(dictionary, key, keyname):
        """helper routine, throws exception if key is not present in dictionary"""
        if key not in dictionary:
            raise fdrtd.server.exceptions.InvalidIdentifier(keyname, key)
        return dictionary[key]

    @staticmethod
    def safe_delete(dictionary, key, keyname):
        """helper routine, throws exception if key is not present in dictionary"""
        if key not in dictionary:
            raise fdrtd.server.exceptions.InvalidIdentifier(keyname, key)
        del dictionary[key]



"""
contains microservice Broadcast
"""

import uuid as _uuid

from fdrtd.server.microservice import Microservice


class Broadcast(Microservice):
    """synchronizes parties through broadcasts"""

    def __init__(self, bus, endpoint):
        super().__init__(bus, endpoint)
        self.storage = {}

    def create(self, tokens):
        """creates a new broadcast"""
        uuid = Broadcast._create_deterministic_uuid(tokens)
        return self.callback(uuid)

    def send(self, callback, message):
        """stores a message to be broadcasted"""
        self.storage[callback] = message

    def receive(self, callback):
        """retrieves a broadcasted message"""
        return self.storage.get(callback, None)

    def delete(self, callback):
        """deletes a broadcast"""
        del self.storage[callback]

    @staticmethod
    def _create_deterministic_uuid(tokens):
        uuid = _uuid.UUID('fede1a7e-0010-4e73-865d-a8e55a223b63')
        uuid = _uuid.uuid5(uuid, 'Broadcast')
        if tokens is not None:
            for token in tokens:
                uuid = _uuid.uuid5(uuid, token)
        return str(uuid)



from django.http import JsonResponse
from django.views.decorators.csrf import csrf_exempt
import json


import fdrtd.webserver.django.fdrtd_app
from fdrtd.server.exceptions import handle_exception


@csrf_exempt
def path_representations(request):
    """route requests to /representations"""
    if request.method == 'GET':
        return list_representations()
    if request.method == 'POST':
        return create_representation(json.loads(request.body))
    if request.method == 'PUT':
        return upload_representation(json.loads(request.body))


@csrf_exempt
def path_representation_uuid(request, uuid):
    """route requests to /representation/<uuid>"""
    if request.method == 'PATCH':
        return call_representation(str(uuid), json.loads(request.body))
    if request.method == 'GET':
        return download_representation(str(uuid))
    if request.method == 'DELETE':
        return release_representation(str(uuid))


@csrf_exempt
def path_representation_uuid_name(request, uuid, name):
    """route requests to /representation/<uuid>/<name>"""
    if request.method == 'GET':
        return create_attribute(str(uuid), name)


def get_bus():
    """get the singleton bus of the server application"""
    return fdrtd.webserver.django.fdrtd_app.fdrtd_bus


def list_representations():
    """list available server-side objects"""
    try:
        response = get_bus().list_representations()
        return JsonResponse({'type': 'list', 'list': response})
    except Exception as exception:
        message, status = handle_exception(exception)
        return JsonResponse({'message': message}, status=status)


def create_representation(body):
    """create a representation"""
    try:
        response = get_bus().create_representation(body)
        return JsonResponse({'type': 'uuid', 'uuid': response})
    except Exception as exception:
        message, status = handle_exception(exception)
        return JsonResponse({'message': message}, status=status)


def upload_representation(body):
    """upload an object, and return its representation"""
    try:
        response = get_bus().upload_representation(body)
        return JsonResponse({'type': 'uuid', 'uuid': response})
    except Exception as exception:
        message, status = handle_exception(exception)
        return JsonResponse({'message': message}, status=status)


def call_representation(representation_uuid, body):
    """call a server-side object"""
    try:
        response = get_bus().call_representation(representation_uuid, body)
        if response is None:
            return JsonResponse({'type': 'none'})
        return JsonResponse({'type': 'uuid', 'uuid': response})
    except Exception as exception:
        message, status = handle_exception(exception)
        return JsonResponse({'message': message}, status=status)


def download_representation(representation_uuid):
    """download a serialized version of a server-side object"""
    try:
        response = get_bus().download_representation(representation_uuid)
        return JsonResponse({'type': 'object', 'object': response})
    except Exception as exception:
        message, status = handle_exception(exception)
        return JsonResponse({'message': message}, status=status)


def release_representation(representation_uuid):
    """release a representation"""
    try:
        get_bus().release_representation(representation_uuid)
        return JsonResponse({'type': 'none'})
    except Exception as exception:
        message, status = handle_exception(exception)
        return JsonResponse({'message': message}, status=status)


def create_attribute(representation_uuid, attribute_name):
    """create a representation of an attribute of a representation"""
    try:
        uuid = get_bus().create_attribute(representation_uuid, attribute_name, public=True)
        return JsonResponse({'type': 'uuid', 'uuid': uuid})
    except Exception as exception:
        message, status = handle_exception(exception)
        return JsonResponse({'message': message}, status=status)



"""
contains the entry points of the API
"""

from flask import current_app

from fdrtd.server.exceptions import handle_exception


def get_bus():
    """get the singleton bus of the server application"""
    with current_app.app_context():
        return current_app.bus


def list_representations():
    """list available server-side objects"""
    try:
        response = get_bus().list_representations()
        return {'type': 'list', 'list': response}, 200  # OK
    except Exception as exception:
        return handle_exception(exception)


def create_representation(body):
    """create a representation"""
    try:
        response = get_bus().create_representation(body)
        return {'type': 'uuid', 'uuid': response}, 200  # OK
    except Exception as exception:
        return handle_exception(exception)


def upload_representation(body):
    """upload an object, and return its representation"""
    try:
        response = get_bus().upload_representation(body)
        return {'type': 'uuid', 'uuid': response}, 200  # OK
    except Exception as exception:
        return handle_exception(exception)


def call_representation(representation_uuid, body):
    """call a server-side object"""
    try:
        response = get_bus().call_representation(representation_uuid, body)
        if response is None:
            return {'type': 'none'}, 200  # OK
        return {'type': 'uuid', 'uuid': response}, 200  # OK
    except Exception as exception:
        return handle_exception(exception)


def download_representation(representation_uuid):
    """download a serialized version of a server-side object"""
    try:
        response = get_bus().download_representation(representation_uuid)
        return {'type': 'object', 'object': response}, 200  # OK
    except Exception as exception:
        return handle_exception(exception)


def release_representation(representation_uuid):
    """release a representation"""
    try:
        get_bus().release_representation(representation_uuid)
        return {'type': 'none'}, 200  # OK
    except Exception as exception:
        return handle_exception(exception)


def create_attribute(representation_uuid, attribute_name):
    """create a representation of an attribute of a representation"""
    try:
        uuid = get_bus().create_attribute(representation_uuid, attribute_name, public=True)
        return {'type': 'uuid', 'uuid': uuid}, 200  # OK
    except Exception as exception:
        return handle_exception(exception)



"""routines to interact with the bus of microservices"""

import uuid as _uuid

import fdrtd.server.callback
import fdrtd.server.exceptions


class Bus:
    """the bus stores representations of microservices and other server-side objects"""

    def __init__(self):
        """initialize the bus"""
        self.microservices = {}
        self.classes = {}
        self.lut_uuid_to_repr = {}

    def set_microservices(self, microservices):
        """set microservices discovered on server startup (internal use only)"""
        self.microservices = microservices

    def set_classes(self, classes):
        """set representations discovered on server startup (internal use only)"""
        self.classes = classes

    def get_argument(self, arg):
        if isinstance(arg, dict):
            if 'representation_uuid' in arg:
                if arg['representation_uuid'] in self.lut_uuid_to_repr:
                    return self.get_argument(self.lut_uuid_to_repr[arg['representation_uuid']])
        return arg

    def get_arguments(self, body):
        args = body.get('args', [])
        kwargs = {}
        for k in body.get('kwargs', {}):
            kwargs[k] = self.get_argument(body['kwargs'][k])
        return args, kwargs

    def list_representations(self):
        """list available server-side objects"""
        result = []
        for _, microservice in self.microservices.items():
            result.append({'identifiers': microservice['identifiers'],
                           'public': list(microservice['public'].keys())})
        return result

    def create_representation(self, body):
        """create a representation"""
        _, requirements = self.get_arguments(body)

        def test(identifiers):
            for requirement in requirements:
                if requirement not in identifiers:
                    return False
                if identifiers[requirement] != requirements[requirement]:
                    return False
            return True

        for uuid, microservice in self.microservices.items():
            if test(microservice['identifiers']):
                if uuid not in self.lut_uuid_to_repr:
                    self.lut_uuid_to_repr[uuid] = microservice
                return uuid

        for uuid, class_descriptor in self.classes.items():
            if test(class_descriptor['identifiers']):
                if uuid not in self.lut_uuid_to_repr:
                    self.lut_uuid_to_repr[uuid] = class_descriptor['class']
                return uuid

        raise fdrtd.server.exceptions.MicroserviceNotFound(requirements)

    def upload_representation(self, body):
        """upload an object, and return its representation"""
        args, _ = self.get_arguments(body)
        uuid = str(_uuid.uuid4())
        self.lut_uuid_to_repr[uuid] = args[0]
        return uuid

    def call_representation(self, representation_uuid, body):
        """call a server-side object"""
        args, kwargs = self.get_arguments(body)
        pointer = self.lut_uuid_to_repr[representation_uuid]
        if isinstance(pointer, dict):
            function = self.lut_uuid_to_repr[pointer['pointer']]
            result = function(*args, **kwargs, callback=pointer['callback'])
        else:
            result = pointer(*args, **kwargs)
        if result is None:
            return None
        uuid = str(_uuid.uuid4())
        self.lut_uuid_to_repr[uuid] = result
        return uuid

    def download_representation(self, representation_uuid):
        """download a serialized version of a server-side object"""
        return self.lut_uuid_to_repr[representation_uuid]

    def release_representation(self, representation_uuid):
        """release a representation"""
        del self.lut_uuid_to_repr[representation_uuid]

    def create_attribute(self, representation_uuid, attribute_name, public=False):
        """create a representation of an attribute of a representation"""

        if representation_uuid in self.microservices:
            instance = self.microservices[representation_uuid]
        else:
            instance = self.lut_uuid_to_repr[representation_uuid]

        if isinstance(instance, fdrtd.server.callback.Callback):
            pointer = {'pointer': self.create_attribute(representation_uuid=instance['handle'],
                                                        attribute_name=attribute_name),
                       'callback': instance['callback']}
        elif isinstance(instance, dict):
            if attribute_name in instance['public']:
                pointer = instance['public'][attribute_name]
            elif public:
                raise fdrtd.server.exceptions.FunctionNotPublic(attribute_name)
            elif attribute_name in instance['private']:
                pointer = instance['private'][attribute_name]
            else:
                raise fdrtd.server.exceptions.FunctionNotFound(attribute_name)
        else:
            try:
                pointer = getattr(instance, attribute_name)
            except KeyError as key_error:
                raise fdrtd.server.exceptions.FunctionNotFound(attribute_name) from key_error

        uuid = str(_uuid.uuid4())
        self.lut_uuid_to_repr[uuid] = pointer
        return uuid



"""routines to discover plugins and their microservices on server startup"""

import uuid as _uuid
import importlib
import pkgutil

import fdrtd.builtins
import fdrtd.plugins


def import_microservice(item, bus, uuid):
    """import a microservice and its public and private functions"""
    instance = item['class'](bus, uuid)
    public = instance.make_public()
    if 'public' in item:
        public.update({function: getattr(instance, function)
                       for function in item['public']})
    private = {}
    for attr in dir(instance):
        if ('__' not in attr) \
                and (callable(getattr(instance, attr))) \
                and (attr not in public):
            private[attr] = getattr(instance, attr)
    return {
        **item,
        'instance': instance,
        'public': public,
        'private': private
    }


def discover_microservices(bus):
    """discover microservices and classes in fdrtd.builtins and fdrtd.plugins"""

    microservices = {}
    classes = {}

    for namespace_package in [fdrtd.builtins, fdrtd.plugins]:
        for _, name, _ in pkgutil.iter_modules(namespace_package.__path__,
                                               namespace_package.__name__ + "."):
            module = importlib.import_module(name)
            try:
                for item in getattr(module, "get_microservices")():
                    uuid = str(_uuid.uuid4())
                    microservices[uuid] = import_microservice(item, bus, uuid)
            except AttributeError:
                pass
            try:
                for item in getattr(module, "get_classes")():
                    uuid = str(_uuid.uuid4())
                    classes[uuid] = {**item, 'instance': item['class']()}
            except AttributeError:
                pass
    return microservices, classes



"""
contains microservice Barrier and its instances
"""

import uuid as _uuid

from fdrtd.server.microservice import Microservice


class Barrier(Microservice):
    """synchronizes parties through barriers"""

    def __init__(self, bus, endpoint):
        super().__init__(bus, endpoint)
        self.storage = {}

    def create(self, tokens):
        """creates a new barrier"""
        uuid = Barrier._create_deterministic_uuid(tokens)
        if uuid not in self.storage:
            self.storage[uuid] = BarrierInstance()
        return self.storage[uuid]

    def delete(self, uuid):
        """deletes a barrier"""
        del self.storage[uuid]

    @staticmethod
    def _create_deterministic_uuid(tokens):
        uuid = _uuid.UUID('fede1a7e-0010-4e73-865d-a8e55a223b63')
        uuid = _uuid.uuid5(uuid, 'Barrier')
        if tokens is not None:
            for token in tokens:
                uuid = _uuid.uuid5(uuid, token)
        return str(uuid)


class BarrierInstance:
    """a synchronization barrier"""

    def __init__(self):
        self.arrivals = set()
        self.departures = set()

    def arrive(self, party):
        """sets a parties state to arrived and resets departures"""
        self.arrivals.add(party)
        self.departures = set()

    def arrived(self):
        """returns the number of arrived parties"""
        return len(self.arrivals)

    def depart(self, party):
        """sets a parties state to departed"""
        self.departures.add(party)

    def departed(self):
        """returns the number of departed parties"""
        return len(self.departures)

    def reset(self):
        """resets arrivals"""
        self.arrivals = set()
